import asyncio
import hashlib
import hmac
import time
from urllib.parse import urlencode, urljoin

import aiohttp
import async_timeout

from .errors import BittrexError


class BittrexAPI:
    """
    https://bittrex.com/home/api -> https://support.bittrex.com/hc/en-us/articles/115003723911
    https://github.com/ericsomdahl/python-bittrex/blob/master/bittrex/bittrex.py
    """
    API_URL = 'https://bittrex.com/api/'

    def __init__(self, api_key=None, api_secret=None, call_rate=1):
        """
        :param api_key: Bittrex api key
        :param api_secret: Bittrex api secret
        :param call_rate: limit requests per seconds
        """
        self.api_key = api_key or ''
        self.api_secret = api_secret or ''
        self.call_interval = 1. / call_rate
        self.timeout = 20
        self._last_call = time.time() - self.call_interval

    async def query(self, path, options=None, authenticate=False, priority=False, version='v1.1'):
        options = options or {}

        if path.startswith('/'):
            path = path[1:]
        url = urljoin(self.API_URL + version + '/', path)

        if authenticate:
            nonce = str(int(time.time() * 1000))
            url = "{url}?apikey={api_key}&nonce={nonce}".format(
                url=url, api_key=self.api_key, nonce=nonce
            )
            if options:
                url += '&' + urlencode(options)
        elif options:
            url += '?' + urlencode(options)

        signature = hmac.new(
            key=self.api_secret.encode(),
            msg=url.encode(),
            digestmod=hashlib.sha512
        ).hexdigest()

        if not priority:
            to_wait = time.time() - self._last_call + self.call_interval
            if to_wait > 0:
                await asyncio.sleep(to_wait)

        try:
            headers = {
                'apisign': signature,
                'Content-Type': 'application/json'
            }
            async with aiohttp.ClientSession(headers=headers) as session:
                with async_timeout.timeout(self.timeout):
                    async with session.get(url) as response:
                        j = await response.json()
                        if not j['success']:
                            raise BittrexError(message=j.get('message', "Unknown error."))
                        return j['result']
        finally:
            self._last_call = time.time()

    def get_markets(self):
        """
        Get the open and available trading markets at Bittrex along with other meta data.
        [{
            "MarketCurrency": "LTC",
            "BaseCurrency": "BTC",
            "MarketCurrencyLong": "Litecoin",
            "BaseCurrencyLong": "Bitcoin",
            "MinTradeSize": 0.01441756,
            "MarketName": "BTC-LTC",
            "IsActive": true,
            "Created": "2014-02-13T00:00:00",
            "Notice": null,
            "IsSponsored": null,
            "LogoUrl": "https://bittrexblobstorage.blob.core.windows.net/public/6defbc41-582d-47a6-bb2e-d0fa88663524.png"
        }]
        """
        return self.query(path='/public/getmarkets')

    def get_currencies(self):
        """
        Get all supported currencies at Bittrex along with other meta data.
        [{
            "Currency": "BTC",
            "CurrencyLong": "Bitcoin",
            "MinConfirmation": 2,
            "TxFee": 0.0005,
            "IsActive": true,
            "CoinType": "BITCOIN",
            "BaseAddress": "1N52wHoVR79PMDishab2XmRHsbekCdGquK",
            "Notice": null
        }]
        """
        return self.query(path='/public/getcurrencies')

    def get_ticker(self, market):
        """
        Get the current tick values for a market.
        {
            "Bid": 0.01702595,
            "Ask": 0.01709242,
            "Last": 0.01702595
        }
        """
        return self.query(path='/public/getticker', options={'market': market})

    def get_market_summaries(self):
        """
        Get the last 24 hour summary of all active markets.
        [{
            "MarketName": "BTC-LTC",
            "High": 0.01717,
            "Low": 0.01664,
            "Volume": 19292.05592121,
            "Last": 0.01709242,
            "BaseVolume": 325.65963883,
            "TimeStamp": "2018-04-23T13:09:54.903",
            "Bid": 0.01702596,
            "Ask": 0.01709242,
            "OpenBuyOrders": 1957,
            "OpenSellOrders": 4016,
            "PrevDay": 0.016837,
            "Created": "2014-02-13T00:00:00"
        }]
        """
        return self.query(path='/public/getmarketsummaries')

    async def get_market_summary(self, market):
        result = await self.query(path='/public/getmarketsummary', options={'market': market})
        """
        Get the last 24 hour summary of a specific market.
        {
            "MarketName": "BTC-LTC",
            "High": 0.01717,
            "Low": 0.01664,
            "Volume": 19298.50773759,
            "Last": 0.017092,
            "BaseVolume": 325.76997876,
            "TimeStamp": "2018-04-23T13:12:20.447",
            "Bid": 0.017092,
            "Ask": 0.01709242,
            "OpenBuyOrders": 1957,
            "OpenSellOrders": 4018,
            "PrevDay": 0.01687339,
            "Created": "2014-02-13T00:00:00"
        }
        """
        if result:
            return result[0]

    def get_order_book(self, market, order_type='both'):
        """
        Retrieve the orderbook for a given market.
        :param order_type: 'buy', 'sell', 'both'
        {
            "buy": [{
                "Quantity": 0.56636808,
                "Rate": 0.01709205
            }],
            "sell": [{
                "Quantity": 67.07309757,
                "Rate": 0.01709242
            }]
        }
        """
        return self.query(
            path='/public/getorderbook',
            options={'market': market, 'type': order_type}
        )

    def get_market_history(self, market):
        """
        Retrieve the latest trades that have occurred for a specific market.
        [{
            "Id": 159594115,
            "TimeStamp": "2018-04-23T12:59:56.333",
            "Quantity": 7.08668072,
            "Price": 0.01702576,
            "Total": 0.12065612,
            "FillType": "PARTIAL_FILL",
            "OrderType": "SELL"
        }, {
            "Id": 159594103,
            "TimeStamp": "2018-04-23T12:59:38.147",
            "Quantity": 1.60041657,
            "Price": 0.01709242,
            "Total": 0.02735499,
            "FillType": "FILL",
            "OrderType": "BUY"
        }]
        """
        return self.query(path='/public/getmarkethistory', options={'market': market})

    def buy_limit(self, market, quantity, rate):
        """
        Place a buy order.
        {
            "uuid": "614c34e4-8d71-11e3-94b5-425861b86ab6"
        }
        """
        return self.query(
            path='/market/buylimit',
            options={'market': market, 'quantity': quantity, 'rate': rate},
            authenticate=True,
            priority=True
        )

    def sell_limit(self, market, quantity, rate):
        """
        Place a sell order.
        {
            "uuid": "614c34e4-8d71-11e3-94b5-425861b86ab6"
        }
        """
        return self.query(
            path='/market/selllimit',
            options={'market': market, 'quantity': quantity, 'rate': rate},
            authenticate=True,
            priority=True
        )

    def cancel_order(self, order_id):
        """
        Cancel a buy or sell order.
        """
        return self.query(
            path='/market/cancel',
            options={'uuid': order_id},
            authenticate=True
        )

    def get_open_orders(self, market=None):
        """
        Get open orders, a market can be specified.
        [{
            "Uuid": null,
            "OrderUuid": "09aa5bb6-8232-41aa-9b78-a5a1093e0211",
            "Exchange": "BTC-LTC",
            "OrderType": "LIMIT_SELL",
            "Quantity": 5.00000000,
            "QuantityRemaining": 5.00000000,
            "Limit": 2.00000000,
            "CommissionPaid": 0.00000000,
            "Price": 0.00000000,
            "PricePerUnit": null,
            "Opened": "2014-07-09T03:55:48.77",
            "Closed": null,
            "CancelInitiated": false,
            "ImmediateOrCancel": false,
            "IsConditional": false,
            "Condition": null,
            "ConditionTarget": null
        }]
        """
        return self.query(
            path='/market/getopenorders',
            options={'market': market} if market else None,
            authenticate=True
        )

    def get_balances(self):
        """
        Retrieve all balances for the account.
        [{
            "Currency": "BSD",
            "Balance": 0.0,
            "Available": 0.0,
            "Pending": 0.0,
            "CryptoAddress": null
        }, {
            "Currency": "BTC",
            "Balance": 6e-08,
            "Available": 6e-08,
            "Pending": 0.0,
            "CryptoAddress": "1JQts7UT3gYTs31p6k5YGj3qjcRQ6XAXsn"
        }]
        """
        return self.query(
            path='/account/getbalances',
            authenticate=True
        )

    def get_balance(self, currency):
        """
        Retrieve balance for specific currency.
        {
            "Currency": "BTC",
            "Balance": 6e-08,
            "Available": 6e-08,
            "Pending": 0.0,
            "CryptoAddress": "1JQts7UT3gYTs31p6k5YGj3qjcRQ6XAXsn"
        }
        """
        return self.query(
            path='/account/getbalance',
            options={'currency': currency},
            authenticate=True
        )

    def get_deposit_address(self, currency):
        """
        Retrieve or generate an address for a specific currency.
        {
            "Currency": "BTC",
            "Address": "1JQts7UT3gYTs31p6k5YGj3qjcRQ6XAXsn"
        }
        """
        return self.query(
            path='/account/getdepositaddress',
            options={'currency': currency},
            authenticate=True
        )

    def withdraw(self, currency, quantity, address):
        """
        Withdraw funds from the account.
        {
            "uuid": "68b5a16c-92de-11e3-ba3b-425861b86ab6"
        }
        """
        return self.query(
            path='/account/withdraw',
            options={'currency': currency, 'quantity': quantity, 'address': address},
            authenticate=True
        )

    def get_order(self, order_id):
        """
        Retrieve a single order by uuid.
        {
            "AccountId": null,
            "OrderUuid": "0cb4c4e4-bdc7-4e13-8c13-430e587d2cc1",
            "Exchange": "BTC-SHLD",
            "Type": "LIMIT_BUY",
            "Quantity": 1000.00000000,
            "QuantityRemaining": 1000.00000000,
            "Limit": 0.00000001,
            "Reserved": 0.00001000,
            "ReserveRemaining": 0.00001000,
            "CommissionReserved": 0.00000002,
            "CommissionReserveRemaining": 0.00000002,
            "CommissionPaid": 0.00000000,
            "Price": 0.00000000,
            "PricePerUnit": null,
            "Opened": "2014-07-13T07:45:46.27",
            "Closed": null,
            "IsOpen": true,
            "Sentinel": "6c454604-22e2-4fb4-892e-179eede20972",
            "CancelInitiated": false,
            "ImmediateOrCancel": false,
            "IsConditional": false,
            "Condition": "NONE",
            "ConditionTarget": null
        }
        """
        return self.query(
            path='/account/getorder',
            options={'uuid': order_id},
            authenticate=True
        )

    def get_order_history(self, market=None):
        """
        Retrieve order history.
        [{
            "OrderUuid": "fd97d393-e9b9-4dd1-9dbf-f288fc72a185",
            "Exchange": "BTC-LTC",
            "TimeStamp": "2014-07-09T04:01:00.667",
            "OrderType": "LIMIT_BUY",
            "Limit": 0.00000001,
            "Quantity": 100000.00000000,
            "QuantityRemaining": 100000.00000000,
            "Commission": 0.00000000,
            "Price": 0.00000000,
            "PricePerUnit": null,
            "IsConditional": false,
            "Condition": null,
            "ConditionTarget": null,
            "ImmediateOrCancel": false
        }]
        """
        return self.query(
            path='/account/getorderhistory',
            options={'market': market} if market else None,
            authenticate=True
        )

    def get_withdrawal_history(self, currency=None):
        """
        Retrieve the account withdrawal history.
        [{
            "PaymentUuid": "88048b42-7a13-4f57-8b7e-109aeeca07d7",
            "Currency": "SAFEX",
            "Amount": 803.7676899,
            "Address": "145J9p6AVjFc2fFV1uyA8d4xweULphyuNv",
            "Opened": "2018-02-20T13:54:41.12",
            "Authorized": true,
            "PendingPayment": false,
            "TxCost": 100.0,
            "TxId": "e1ded8356d2855716ba99ae6b8cbd2c4220a8df15dd37fd7eb29a76dd7a0b1d1",
            "Canceled": false,
            "InvalidAddress": false
        }]
        """
        return self.query(
            path='/account/getwithdrawalhistory',
            options={'currency': currency} if currency else None,
            authenticate=True
        )

    def get_deposit_history(self, currency=None):
        """
        Retrieve the account deposit history.
        [{
            "Id": 41565639,
            "Amount": 0.008,
            "Currency": "BTC",
            "Confirmations": 3,
            "LastUpdated": "2017-11-20T16:40:30.6",
            "TxId": "abfec55561b5440b28784dc4b152635c05139f33faec090a3d8e18a8d2c75eec",
            "CryptoAddress": "1JQts7UT3gYTs31p6k5YGj3qjcRQ6XAXsn"
        }]
        """
        return self.query(
            path='/account/getdeposithistory',
            options={'currency': currency} if currency else None,
            authenticate=True
        )

    # v2.0

    def get_wallet_health(self):
        """
        View wallets health.
        [{
            "Health": {
                "Currency": "BTC",
                "DepositQueueDepth": 0,
                "WithdrawQueueDepth": 24,
                "BlockHeight": 519583,
                "WalletBalance": 0.0,
                "WalletConnections": 8,
                "MinutesSinceBHUpdated": 2,
                "LastChecked": "2018-04-23T13:50:11.827",
                "IsActive": true
            },
            "Currency": {
                "Currency": "BTC",
                "CurrencyLong": "Bitcoin",
                "MinConfirmation": 2,
                "TxFee": 0.0005,
                "IsActive": true,
                "CoinType": "BITCOIN",
                "BaseAddress": "1N52wHoVR79PMDishab2XmRHsbekCdGquK",
                "Notice": null
            }
        }]
        """
        return self.query(
            path='/pub/Currencies/GetWalletHealth',
            version='v2.0'
        )

    def get_pending_withdrawals(self, currency=None):
        """
        Get the account pending withdrawals.
        """
        return self.query(
            path='/key/balance/getpendingwithdrawals',
            options={'currency': currency} if currency else None,
            authenticate=True,
            version='v2.0'
        )

    def get_pending_deposits(self, currency=None):
        """
        Get the account pending deposits.
        """
        return self.query(
            path='/key/balance/getpendingdeposits',
            options={'currency': currency} if currency else None,
            authenticate=True,
            version='v2.0'
        )

    def get_candles(self, market, tick_interval):
        """
        Get tick candles for market.
        Intervals: oneMin, fiveMin, hour, day
        [{
            "O": 0.017059,
            "H": 0.01712003,
            "L": 0.017059,
            "C": 0.017059,
            "V": 49.10766337,
            "T": "2018-04-23T14:07:00",
            "BV": 0.83816494
        }]
        """
        return self.query(
            path='/pub/market/GetTicks',
            options={'marketName': market, 'tickInterval': tick_interval},
            version='v2.0'
        )

    async def get_latest_candle(self, market, tick_interval):
        """
        Get the latest candle for the market.
        {
            "O": 0.017125,
            "H": 0.017125,
            "L": 0.01706,
            "C": 0.017125,
            "V": 2.35065452,
            "T": "2018-04-23T14:09:00",
            "BV": 0.04018997
        }
        """
        result = await self.query(
            path='/pub/market/GetLatestTick',
            options={'marketName': market, 'tickInterval': tick_interval},
            version='v2.0'
        )
        if result:
            return result[0]



from unittest import TestCase

from aiobittrex.socket import BittrexSocket


class ReplaceKeysTestCase(TestCase):

    def test_replace_keys(self):
        inp = {
            "BTC-TRX": {
                "M": None,
                "N": 10647,
                "Z": [{
                    "Q": 242.46304653,
                    "R": 8.67e-06
                }, {
                    "Q": 417459.77306401,
                    "R": 8.64e-06
                }]
            }
        }

        out = {
            'BTC-TRX': {
                'buys': [{
                    'quantity': 242.46304653,
                    'rate': 8.67e-06
                }, {
                    'quantity': 417459.77306401,
                    'rate': 8.64e-06
                }],
                'market_name': None,
                'nonce': 10647
            }
        }

        self.assertEqual(BittrexSocket.replace_keys(inp), out)



import asyncio
import hashlib
import hmac
import json
import logging
import time
from base64 import b64decode
from urllib.parse import urlencode
from zlib import decompress, MAX_WBITS

import aiohttp

from .errors import BittrexError

logger = logging.getLogger(__name__)


class BittrexSocket:
    """
    https://bittrex.github.io/
    """
    SOCKET_URL = 'https://socket.bittrex.com/signalr/'
    SOCKET_HUB = 'c2'

    KEYS = {
        'A': 'ask',
        'a': 'available',
        'B': 'bid',
        'b': 'balance',
        'C': 'closed',
        'c': 'currency',
        'CI': 'cancel_initiated',
        'D': 'deltas',
        'd': 'delta',
        'DT': 'order_delta_type',
        'E': 'exchange',
        'e': 'exchange_delta_type',
        'F': 'fill_type',
        'FI': 'fill_id',
        'f': 'fills',
        'G': 'open_buy_orders',
        'g': 'open_sell_orders',
        'H': 'high',
        'h': 'auto_sell',
        'I': 'id',
        'i': 'is_open',
        'J': 'condition',
        'j': 'condition_target',
        'K': 'immediate_or_cancel',
        'k': 'is_conditional',
        'L': 'low',
        'l': 'last',
        'M': 'market_name',
        'm': 'base_volume',
        'N': 'nonce',
        'n': 'commission_paid',
        'O': 'orders',
        'o': 'order',
        'OT': 'order_type',
        'OU': 'order_uuid',
        'P': 'price',
        'p': 'crypto_address',
        'PD': 'prev_day',
        'PU': 'price_per_unit',
        'Q': 'quantity',
        'q': 'quantity_remaining',
        'R': 'rate',
        'r': 'requested',
        'S': 'sells',
        's': 'summaries',
        'T': 'time_stamp',
        't': 'total',
        'TY': 'type',
        'U': 'uuid',
        'u': 'updated',
        'V': 'volume',
        'W': 'account_id',
        'w': 'account_uuid',
        'X': 'limit',
        'x': 'created',
        'Y': 'opened',
        'y': 'state',
        'Z': 'buys',
        'z': 'pending'
    }

    def __init__(self, api_key=None, api_secret=None, loop=None):
        self.api_key = api_key
        self.api_secret = api_secret
        self._socket_url = None
        self._loop = loop or asyncio.get_event_loop()
        self._session = aiohttp.ClientSession(loop=loop)

    def __del__(self):
        self._loop.run_until_complete(self._session.close())

    @staticmethod
    def _decode(message):
        try:
            deflated_msg = decompress(b64decode(message, validate=True), -MAX_WBITS)
        except SyntaxError:
            deflated_msg = decompress(b64decode(message, validate=True))
        return json.loads(deflated_msg.decode())

    @classmethod
    def replace_keys(cls, d):
        if not isinstance(d, dict):
            return d
        result = {}
        for key, value in d.items():
            key = cls.KEYS.get(key, key)
            if isinstance(value, dict):
                result[key] = cls.replace_keys(value)
            elif isinstance(value, list):
                result[key] = [cls.replace_keys(v) for v in value]
            else:
                result[key] = value
        return result

    async def _get_socket_url(self):
        if self._socket_url is None:
            conn_data = json.dumps([{'name': self.SOCKET_HUB}])
            url = self.SOCKET_URL + 'negotiate' + '?' + urlencode({
                'clientProtocol': '1.5',
                'connectionData': conn_data,
                '_': round(time.time() * 1000)
            })

            async with self._session.get(url) as r:
                socket_conf = await r.json()

            self._socket_url = self.SOCKET_URL.replace('https', 'wss') + 'connect' + '?' + urlencode({
                'transport': 'webSockets',
                'clientProtocol': socket_conf['ProtocolVersion'],
                'connectionToken': socket_conf['ConnectionToken'],
                'connectionData': conn_data,
                'tid': 3
            })

        return self._socket_url

    async def create_ws(self):
        return await self._session.ws_connect(await self._get_socket_url())

    async def _listen(self, endpoint, messages, ws=None):
        ws = ws or await self.create_ws()

        for n, m in enumerate(messages, start=1):
            await ws.send_str(json.dumps({
                'H': self.SOCKET_HUB,
                'M': endpoint,
                'A': m,
                'I': n
            }))

        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                decoded_message = json.loads(msg.data)
                if 'E' in decoded_message:
                    raise BittrexError(message=decoded_message['E'])
                yield decoded_message
            elif msg.type == aiohttp.WSMsgType.closed:
                break
            elif msg.type == aiohttp.WSMsgType.error:
                break
            else:
                logger.warning("Message: {}".format(msg.type))

    async def _get_auth_context(self, ws):
        async for m in self._listen(endpoint='GetAuthContext', messages=[[self.api_key]], ws=ws):
            if 'R' in m:
                return m['R']

    async def listen_account(self, ws=None):
        """
        Listen to account balance and orders updates.

        callbacks:
        uB - balance delta
        uO - order delta
        """
        challenge = await self._get_auth_context(ws)
        signature = hmac.new(
            key=self.api_secret.encode(),
            msg=challenge.encode(),
            digestmod=hashlib.sha512
        ).hexdigest()
        async for m in self._listen(endpoint='Authenticate', messages=[[self.api_key, signature]], ws=ws):
            if 'R' in m:
                assert m['R']

            for row in m.get('M') or []:
                if row['M'] not in ('uB', 'uO'):
                    continue
                for a in row['A']:
                    yield self.replace_keys(self._decode(a))

    async def get_market(self, markets):
        """
        {
            "BTC-TRX": {
                "market_name": null,
                "nonce": 11333,
                "buys": [{
                    "quantity": 428996.57288094,
                    "rate": 8.65e-06
                }],
                "sells": [{
                    "quantity": 91814.92314615,
                    "rate": 8.66e-06
                }],
                "fills": [{
                    "id": 5020055,
                    "time_stamp": 1524904823903,
                    "quantity": 34413.0,
                    "price": 8.66e-06,
                    "total": 0.29801658,
                    "fill_type": "FILL",
                    "order_type": "BUY"
                }]
            }
        }
        """
        result = {}
        async for m in self._listen(endpoint='QueryExchangeState', messages=[[m] for m in markets]):
            if 'R' not in m:
                continue
            i = int(m['I'])
            result[markets[i - 1]] = self.replace_keys(self._decode(m['R']))
            if len(result) >= len(markets):
                break
        return result

    async def listen_market(self, markets, ws=None):
        """
        Listen to market updates.

        callbacks:
        - uE - market delta

        {
            "market_name": "BTC-TRX",
            "nonce": 11919,
            "buys": [],
            "sells": [{
                "type": 2,
                "rate": 8.7e-06,
                "quantity": 197473.52148216
            }],
            "fills": [{
                "order_type": "BUY",
                "rate": 8.7e-06,
                "quantity": 28376.84449489,
                "time_stamp": 1524905878547
            }]
        }
        """
        async for m in self._listen(endpoint='SubscribeToExchangeDeltas', messages=[[m] for m in markets], ws=ws):
            for row in m.get('M') or []:
                if row['M'] != 'uE':
                    continue
                for a in row['A']:
                    yield self.replace_keys(self._decode(a))

    async def get_summary(self):
        """
        {
            "nonce": 5108,
            "summaries": [{
                "market_name": "BTC-ADA",
                "high": 3.388e-05,
                "low": 3.116e-05,
                "volume": 45482116.6444527,
                "last": 3.337e-05,
                "base_volume": 1481.80378307,
                "time_stamp": 1524907023543,
                "bid": 3.333e-05,
                "ask": 3.337e-05,
                "open_buy_orders": 5195,
                "open_sell_orders": 15219,
                "prev_day": 3.118e-05,
                "created": 1506668518873
            }]
        }
        """
        async for m in self._listen(endpoint='QuerySummaryState', messages=['']):
            if 'R' in m:
                return self.replace_keys(self._decode(m['R']))

    async def listen_summary_light(self, ws=None):
        """
        callbacks:
        - uL - light summary delta

        {
            "deltas": [{
                "market_name": "BTC-ADT",
                "last": 7.37e-06,
                "base_volume": 118.05
            }]
        }
        """
        async for m in self._listen(endpoint='SubscribeToSummaryLiteDeltas', messages=[''], ws=ws):
            for row in m.get('M') or []:
                if row['M'] != 'uL':
                    continue
                for a in row['A']:
                    yield self.replace_keys(self._decode(a))

    async def listen_summary(self, ws=None):
        """
        callbacks:
        - uS - summary delta

        {
            "nonce": 5069,
            "deltas": [{
                "market_name": "BTC-ETH",
                "high": 0.07371794,
                "low": 0.071695,
                "volume": 9535.44197173,
                "last": 0.07318011,
                "base_volume": 695.21677418,
                "time_stamp": 1524907827823,
                "bid": 0.07318011,
                "ask": 0.07346991,
                "open_buy_orders": 4428,
                "open_sell_orders": 3860,
                "prev_day": 0.07188519,
                "created": 1439542944817
            }]
        }
        """
        async for m in self._listen(endpoint='SubscribeToSummaryDeltas', messages=[''], ws=ws):
            for row in m.get('M') or []:
                if row['M'] != 'uS':
                    continue
                for a in row['A']:
                    yield self.replace_keys(self._decode(a))
